# from paxd.process import print_exc
# from multiprocessing import Pipe, Process
# from Queue import Empty
# 
# def start_appman(name, command_queue, read_q, write_q):
#     AppManager(name, command_queue, read_q, write_q).start()
#     
# class AppManager(object):
#     def __init__(self, name, command_queue, read_q, write_q):
#         self.name = name
#         self.command_queue = command_queue
#         self.read_q = read_q
#         self.write_q = write_q
#     
#     def start(self):
#         while True:
#             with print_exc:
#                 obj = self.read_q.get()
#                 self.handle(obj)
#     
#     def handle(self, obj):
#         # Validate Command
#         if obj == 'APPLIST':
#             self.write_q.put({
#                 'applications' : []
#             })
#             return
#         return {'error' : 'unknown command'}
#         # 
#         # # Execute Command
#         # self.command_queue.put((name, 'APPLIST'))
#         # if controller_pipe.poll(0.8):
#         #     httpserver_pipe.send(controller_pipe.recv())
#         #     return
#         # httpserver_pipe.send('TIMEOUT')
#         # 



from multiprocessing import Process, Pipe, Queue
from Queue import Empty
from signal import signal, SIGINT, SIGKILL, SIGTERM
from threading import RLock, Thread
import os
import sys
import pickle
import json
from redis import Redis

from paxd.app.application import Application
from paxd.process import print_exc
from paxd.server.webstats import start_webui
from paxd.app.msg import BaseRequest, BaseResponse

PROJECT_PATH = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))

class Controller(object):
    command_queue = '_paxd.controller'
    webui_queue = '_paxd.web_ui'
    def __init__(self, instance_id):
        self.instance_id = instance_id
        self.lock = RLock()
        self.processes = []
        self.threads = []
        self.applications = []
    
    def run(self):
        try:
            print 'STARTING UP CONTROLLER.  pid: ', os.getpid()
            # Internal Applications
            self.start_process(start_webui, self.webui_queue)
            
            # Start "apps"
            self.configure()
            
            # Wait For Commands
            self.handle_commands()
        except Exception, e:
            print 'TERMINATING', e
            import sys
            for process in self.processes:
                process.terminate()
            raise
    
    def configure(self):
        redis = Redis('localhost')
        config = redis.hget('_paxd.controllers.config', self.instance_id)
        if not config:
            self.load_app(PROJECT_PATH, 'paxd.webuiapp.webui.webui_app', self.webui_queue, name="paxd admin/api", unstoppable=True)
            return
        config = json.loads(config)
        for app in config['applications']:
            application = self.load_app(app['path'], app['entry'], app['queue'], environment=app['environment'], name=app['name'], unstoppable=app['unstoppable'], id=app['id'])
            if app['status'] == 'stopped' or app['status'] == 'paused':
                application.pause()
    
    def handle_commands(self):
        thread = Thread(target=command_thread, args=(self,))
        thread.daemon = True
        thread.start()
        self.threads.append(thread)
        while True:
            try:
                thread.join(60)
            except KeyboardInterrupt:
                print 'CONTROLLER EXITING'
                sys.exit(0)
    
    def load_app(self, path, entry, queue, environment=None, logger=None, name=None, unstoppable=False, id=None):
        if not name:
            name = entry
        application = Application(self.instance_id, name, path, entry, queue, logger, environment=environment, unstoppable=unstoppable, id=id)
        with self.lock:
            self.applications.append(application)
        application.run()
        return application
    
    def load_auto(self, path):
        with open(os.path.join(path, 'pax.conf')) as conf_file:
            conf = json.loads(conf_file.read())
        
        environment = {}
        for k, v in conf.get('ENVIRONMENT',{}).iteritems():
            environment[k] = v
        print conf.get('ENVIRONMENT',{})
        apps = []
        for entry, appconf in conf['APPS'].iteritems():
            apps.append(self.load_app(path, entry, entry, environment=environment))
        return apps
    
    def start_process(self, fun, name):
        process = Process(target=fun, args=(name,))
        process.start()
        print 'STARTING INTERNAL PROCESS', name, process.pid
        self.processes.append(process)

def command_thread(controller):
    redis = Redis('localhost')
    while True:
        with print_exc:
            _, value = redis.blpop(controller.command_queue)
            request = BaseRequest.from_serialized(value)
            
            resp_data = handle_command(controller, redis, request.deserialize(request.args))
            
            response = request.response_class(request.response_queue, 'SUCCESS', resp_data)
            pipe = redis.pipeline()
            q = request.response_queue
            ser = response.serialize_resp()
            pipe.lpush(q, ser)
            pipe.expire(request.response_queue, request.timeout)
            # pipe.lrem(self.active_queue, 1, request.raw)
            pipe.execute()
            save_settings(redis, controller)

def save_settings(redis, controller):
    pipe = redis.pipeline()
    apps = []
    with controller.lock:
        for app in controller.applications:
            apps.append(app.json_info)
    redis.hset('_paxd.controllers.config', controller.instance_id, json.dumps({
        'applications' : apps,
    }))

def handle_command(controller, redis, rdata):
    command = rdata['command']
    if command == 'APPLIST':
        return handle_list(controller, redis, rdata)
    elif command == 'LOAD_APP':
        return handle_load_app(controller, redis, rdata)
    elif command == 'LOAD_AUTO':
        return handle_load_auto(controller, redis, rdata)
    elif command == 'UNLOAD':
        return handle_unload(controller, redis, rdata)
    elif command == 'PAUSE' or command == 'UNPAUSE':
        return handle_pause_unpause(controller, redis, rdata)
    elif command == 'GET_FAILED':
        return handle_get_failed(controller, redis, rdata)
    return 'BAD COMMAND'

def handle_list(controller, redis, rdata):
    ret = []
    with controller.lock:
        for app in controller.applications:
            json_info = app.json_info
            json_info.update(**app.queue_info(redis))
            ret.append(json_info)
    return ret

def handle_get_failed(controller, redis, rdata):
    id = rdata['id']
    with controller.lock:
        for app in controller.applications:
            if app.id == id:
                return { 'status' : 'SUCCESS', 'failed' : app.failed_items, 'app' : app.json_info }
    return { 'status' : 'FAILURE', 'reason' : 'NO SUCH APP ID' }
    

def handle_load_app(controller, redis, rdata):
    apps = controller.load_app(rdata['path'], rdata['entrypoint'], rdata['queue'], None)
    return { 'status' : 'SUCCESS', 'apps' : [app.json_info] }

def handle_load_auto(controller, redis, rdata):
    apps = controller.load_auto(rdata['path'])
    return { 'status' : 'SUCCESS', 'apps' : [app.json_info for app in apps] }


def handle_unload(controller, redis, rdata):
    id = rdata['id']
    with controller.lock:
        for app in controller.applications:
            if app.id == id:
                app.quit()
                break
        else:
            return { 'status' : 'FAILURE', 'reason' : 'NO SUCH APP ID' }
        controller.applications = [a for a in controller.applications if a.id != id]
        return { 'status' : 'SUCCESS', 'app' : app.json_info }

def handle_pause_unpause(controller, redis, rdata):
    id = rdata['id']
    with controller.lock:
        for app in controller.applications:
            if app.id == id:
                if rdata['command'] == 'PAUSE':
                    app.pause()
                else:
                    app.unpause()
                return { 'status' : 'SUCCESS', 'app' : app.json_info }
    return { 'status' : 'FAILURE', 'reason' : 'NO SUCH APP ID' }




import uuid
import traceback
import json
import pickle
from paxd.app.http import Response
from paxd.app.msg import PRequest
from redis import Redis
from paxd.webuiapp import html

def webui_app(request):
    try:
        response = handle_request(request)
    except:
        response = Response(500, 'Server Error %s' % traceback.format_exc())
    return response

def handle_request(request):
    if request.path == '/':
        return Response(200, get_apps())
    elif request.path.startswith('/load'):
        return Response(200, load_app(request))
    elif request.path.startswith('/unload'):
        return Response(200, unload_app(request))
    elif request.path.startswith('/failed'):
        return Response(200, get_failed(request))
    elif request.path.startswith('/pause'):
        return Response(200, pause_app(request))
    elif request.path.startswith('/unpause'):
        return Response(200, unpause_app(request))
    return Response(404, 'Not Found')

controller_queue = '_paxd.controller'
def get_apps():
    redis = Redis('localhost')
    req = PRequest(redis, controller_queue, {
            'command' : 'APPLIST'
        }, timeout=2)
    promise = req.send()
    resp = promise.get()
    if not resp:
        return 'NO RESPONSE'
    
    return html.header() + html.applications(resp) + html.footer()

def load_app(request):
    redis = Redis('localhost')
    single_app = request.GET.get('queue') and request.GET.get('entry')
    autoconf = request.GET.get('autoconf')
    
    assert autoconf == 'file' or single_app, (single_app, autoconf)
    
    if single_app:
        promise = PRequest(redis, controller_queue, {
            'command' : 'LOAD_APP',
            'path' : request.GET['path'], 
            'entrypoint': request.GET.get('entry'), 
            'queue' : request.GET.get('queue'),
        })
    else:
        promise = PRequest(redis, controller_queue, {
            'command' : 'LOAD_AUTO',
            'path' : request.GET['path'], 
            'autoconfigure' : request.GET.get('autoconfigure') == 'true',
        })
        


    resp = promise.send().get()
    if not resp:
        return 'NO RESPONSE'
    return json.dumps(resp)

def get_failed(request):
    redis = Redis('localhost')
    promise = PRequest(redis, controller_queue, {
        'command' : 'GET_FAILED',
        'id' : request.GET['id'], 
    }, timeout=10)
    resp = promise.send().get()
    if not resp:
        return 'NO RESPONSE'
    return html.header() + html.failed(resp) + html.footer()

def unload_app(request):
    redis = Redis('localhost')
    promise = PRequest(redis, controller_queue, {
        'command' : 'UNLOAD',
        'id' : request.GET['id'], 
    }, timeout=10)
    resp = promise.send().get()
    if not resp:
        return 'NO RESPONSE'
    return json.dumps(resp)

def pause_app(request):
    redis = Redis('localhost')
    promise = PRequest(redis, controller_queue, {
        'command' : 'PAUSE',
        'id' : request.GET['id'], 
    }, timeout=4)
    resp = promise.send().get()
    if not resp:
        return 'NO RESPONSE'
    return json.dumps(resp)

def unpause_app(request):
    redis = Redis('localhost')
    promise = PRequest(redis, controller_queue, {
        'command' : 'UNPAUSE',
        'id' : request.GET['id'], 
    }, timeout=4)
    resp = promise.send().get()
    if not resp:
        return 'NO RESPONSE'
    return json.dumps(resp)



import os
from StringIO import StringIO

def header():
    return u'''
<html>
<head>
<title>Pax Daemonica</title>
    <script type="text/javascript" src="https://ajax.googleapis.com/ajax/libs/jquery/1.6.1/jquery.min.js"></script>
    %s
    %s
</head>
<body>
    ''' % (styles(), js())

def footer():
    return '''
    </body>
    </html>
    
    '''

def applications(apps):
    ret = u'''<table>'''
    ret += app_header()
    for app in apps:
        ret += application(app)
    return ret + u'</table>'

def app_header():
    return u'''<tr>
        <th>Q       </th>
        <th>Q:a     </th>
        <th>Q:f     </th>
        <th>Q:s     </th>
        <th>controls</th>
        <th>status  </th>
        <th>queue   </th>
        <th>name    </th>
        <th>entry   </th>
        <th>path    </th>
        <th>id      </th>
        </tr>
    '''

def application(app):
    id_query = u'?id=' + app['id']
    unstoppable = ''
    if app['unstoppable']:
        unstoppable = 'unstoppable'
    q = app.get('queued')
    qa = app.get('active')
    qf = app.get('failed')
    qs = app.get('success')
    return u'''
        <tr class="{unstoppable}">
            <td class="path">{q}</td>
            <td class="path">{qa}</td>
            <td class="path">
                <a class="failed" href="/failed{id_query}">{qf}</a>
            </td>
            <td class="path">{qs}</td>
            <td class="actions">
                <a class="action start" data-action="start" href="/unpause{id_query}">Start</a> |
                <a class="action stop" data-action="stop" href="/pause{id_query}">Stop</a> |
                <a class="action unload" data-action="unload" href="/unload{id_query}">Remove</a>
            </td>
            <td class="status">{app[status]}</td>
            <td class="queue">{app[queue]}</td>
            <td class="name">{app[name]}</td>
            <td class="entry">{app[entry]}</td>
            <td class="path">{app[path]}</td>
            <td class="id">{app[id]}</td>

        </tr>
    '''.format(**vars())

def failed(resp):
    app = resp['app']
    failed = resp['failed']
    s = StringIO()
    name = app['name']
    id = app['id']
    s.write(u'''
        <h2>{name} ({id})</h2>
        <table class="error">
    '''.format(**vars()))
    for f in failed:
        status = f['response']['status']
        time = f.get('time')
        type = f['response'].get('error_type')
        value = f['response'].get('error_value')
        traceback = f['response'].get('traceback')
        s.write(''' 
            <tr>
                <td>{status}</td>
                <td>{time}</td>
                <td>{type}</td>
                <td>{value}</td>
                <td><pre>{traceback}</pre></td>
            </tr>
        '''.format(**vars()))
    s.write('''
        </table>
    ''')
    return s.getvalue()

def load_file(name):
    file = os.path.join(os.path.dirname(__file__), name)
    with open(file) as f:
        return f.read()

def styles():
    return '''<style>%s</style>''' % load_file('webuiapp.css')

def js():
    return ''' <script>%s\n%s</script>''' % (load_file('json2.js'), load_file('webuiapp.js'))





import time
from paxd.client import task

@task(queue='testapp.app.exclaimer')
def exclaimer(thing):
    time.sleep(0.2)
    return str(thing) + '!!!!!!!'

@task(queue='testapp.app.error_test')
def error_test():
    raise Exception('BAD BAD BAD')
